"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const debug_1 = __importDefault(require("debug"));
const utils_1 = require("./utils");
const lodash_isequal_1 = __importDefault(require("lodash.isequal"));
const url_1 = require("url");
const querystring_1 = require("querystring");
const path_to_regexp_1 = require("path-to-regexp");
const debug = debug_1.default('mockiavelli:mock');
const GET_REQUEST_TIMEOUT = 100;
let debugId = 1;
class Mock {
    constructor(matcher, response, options = {}) {
        this.requests = [];
        this.debugId = debugId++;
        this.options = {
            priority: 0,
            once: false,
        };
        this.matcher = this.parseMatcher(matcher);
        this.response = response;
        this.options = Object.assign(Object.assign({}, this.options), options);
        this.debug('+', `created mock: method=${matcher.method} url=${matcher.url}`);
    }
    debug(symbol, message) {
        debug(`${symbol} (${this.debugId}) ${message} `);
    }
    debugMiss(reason, requestValue, matcherValue) {
        this.debug(`·`, `${reason} not matched: mock=${matcherValue} req=${requestValue} `);
    }
    async waitForRequest(index = 0) {
        try {
            await utils_1.waitFor(() => Boolean(this.requests[index]), GET_REQUEST_TIMEOUT);
        }
        catch (e) {
            if (e instanceof utils_1.TimeoutError) {
                if (this.requests.length === 0 && index === 0) {
                    throw new Error(`No request matching mock [${this.prettyPrint()}] found`);
                }
                else {
                    throw new Error(`${utils_1.nth(index + 1)} request matching mock [${this.prettyPrint()}] was not found`);
                }
            }
            throw e;
        }
        return this.requests[index];
    }
    async waitForRequestsCount(count) {
        try {
            await utils_1.waitFor(() => this.requests.length === count, GET_REQUEST_TIMEOUT);
        }
        catch (e) {
            if (e instanceof utils_1.TimeoutError) {
                throw new Error(`Expected ${count} requests to match mock [${this.prettyPrint()}]. Instead ${this.requests.length} request were matched.`);
            }
            throw e;
        }
    }
    getResponseForRequest(request) {
        if (this.options.once && this.requests.length > 0) {
            this.debug('once', 'Request already matched');
            return null;
        }
        const matchedRequest = this.getRequestMatch(request);
        if (matchedRequest === null) {
            return null;
        }
        this.requests.push(matchedRequest);
        const response = typeof this.response === 'function'
            ? this.response(matchedRequest)
            : this.response;
        return response;
    }
    getRequestMatch(request) {
        if (request.method !== this.matcher.method) {
            this.debugMiss('method', request.method, this.matcher.method);
            return null;
        }
        const matcherOrigin = this.matcher.hostname || request.sourceOrigin;
        if (matcherOrigin !== request.hostname) {
            this.debugMiss('url', request.hostname || `Request origin missing`, matcherOrigin || `Matcher origin missing`);
            return null;
        }
        const pathMatch = this.matcher.pathMatch(request.path);
        if (!pathMatch) {
            this.debugMiss('url', request.path || `Request path missing`, this.matcher.path || `Matcher path missing`);
            return null;
        }
        if (!this.requestParamsMatch(request.query, this.matcher.query)) {
            this.debugMiss('query', JSON.stringify(request.query), JSON.stringify(this.matcher.query));
            return null;
        }
        if (this.matcher.body && !lodash_isequal_1.default(request.body, this.matcher.body)) {
            this.debugMiss('body', JSON.stringify(request.body), JSON.stringify(this.matcher.body));
            return null;
        }
        this.debug('=', `matched mock`);
        return {
            url: request.url,
            path: request.path,
            query: request.query,
            method: request.method,
            body: request.body,
            headers: request.headers,
            params: pathMatch.params,
        };
    }
    parseMatcher(matcher) {
        // TODO find a better alternative for url.parse
        const { protocol, host, pathname, query } = url_1.parse(matcher.url, true);
        const hasHostname = protocol && host;
        return {
            method: matcher.method,
            hostname: hasHostname ? `${protocol}//${host}` : undefined,
            query: matcher.query ? matcher.query : query,
            path: pathname !== null && pathname !== void 0 ? pathname : '',
            pathMatch: path_to_regexp_1.match(pathname !== null && pathname !== void 0 ? pathname : ''),
            body: matcher.body,
        };
    }
    requestParamsMatch(requestQuery, filterQuery) {
        return Object.keys(filterQuery).every((key) => {
            return (key in requestQuery &&
                lodash_isequal_1.default(filterQuery[key], requestQuery[key]));
        });
    }
    prettyPrint() {
        const qs = querystring_1.stringify(this.matcher.query);
        return `(${this.debugId}) ${this.matcher.method || 'HTTP'} ${this.matcher.path + (qs ? '?' + qs : '')}`;
    }
}
exports.Mock = Mock;
//# sourceMappingURL=mock.js.map