"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const debug_1 = __importDefault(require("debug"));
const mock_1 = require("./mock");
const utils_1 = require("./utils");
const BrowserControllerFactory_1 = require("./controllers/BrowserControllerFactory");
const debug = debug_1.default('mockiavelli:main');
const interceptedTypes = ['xhr', 'fetch'];
class Mockiavelli {
    constructor(options = {}) {
        this.mocks = [];
        this.onRequest = async (request, respond, skip) => {
            debug(`> req: ${utils_1.printRequest(request)} `);
            // Handle preflight requests
            if (request.method === 'OPTIONS') {
                return await respond({
                    status: 204,
                    headers: utils_1.sanitizeHeaders(utils_1.getCorsHeaders(request)),
                });
            }
            for (const mock of this.mocks) {
                const response = mock.getResponseForRequest(request);
                if (response) {
                    const status = response.status || 200;
                    // Convert response body to Buffer.
                    // A bug in puppeteer causes stalled response when body is equal to "" or undefined.
                    // Providing response as Buffer fixes it.
                    let body;
                    let contentType;
                    if (typeof response.body === 'string') {
                        body = Buffer.from(response.body);
                    }
                    else if (response.body === undefined ||
                        response.body === null) {
                        body = Buffer.alloc(0);
                    }
                    else {
                        try {
                            body = Buffer.from(JSON.stringify(response.body));
                            contentType = 'application/json; charset=utf-8';
                        }
                        catch (e) {
                            // Response body in either not JSON-serializable or something else
                            // that cannot be handled. In this case we throw an error
                            console.error('Could not serialize response body', e);
                            throw e;
                        }
                    }
                    // Set default value of Content-Type header
                    const headers = utils_1.sanitizeHeaders(Object.assign(Object.assign({ 'content-length': String(body.length), 'content-type': contentType }, utils_1.getCorsHeaders(request)), response.headers));
                    try {
                        await respond({
                            status,
                            headers,
                            body,
                        });
                        debug(`< res: ${utils_1.printResponse(status, headers, body)}`);
                        return;
                    }
                    catch (e) {
                        console.error(`Failed to reply with mocked response for ${utils_1.printRequest(request)}`);
                        console.error(e);
                        throw e;
                    }
                }
            }
            const should404 = interceptedTypes.includes(request.type);
            // Request was not matched - log error and return 404
            if (should404) {
                debug(`< res: status=404`);
                console.error(`Mock not found for request: ${utils_1.printRequest(request)}`);
                return respond({
                    status: 404,
                    body: 'No mock provided for request',
                });
            }
            // Do not intercept non xhr/fetch requests
            debug(`< res: continue`);
            try {
                return await skip();
            }
            catch (e) {
                console.error(e);
                // Request could be already handled so ignore this error
                return;
            }
        };
        if (options.debug) {
            debug_1.default.enable('mockiavelli:*');
        }
        debug('Initialized');
    }
    static async setup(page, options = {}) {
        const instance = new Mockiavelli(options);
        const controller = BrowserControllerFactory_1.BrowserControllerFactory.createForPage(page);
        await instance.activate(controller);
        return instance;
    }
    async activate(controller) {
        await controller.startInterception(this.onRequest);
    }
    mock(matcher, response, options) {
        const mock = new mock_1.Mock(matcher, response, Object.assign({}, options));
        utils_1.addMockByPriority(this.mocks, mock);
        return mock;
    }
    mockGET(matcher, response, options) {
        return this.mock(utils_1.createRequestMatcher(matcher, 'GET'), response, options);
    }
    mockPOST(matcher, response, options) {
        return this.mock(utils_1.createRequestMatcher(matcher, 'POST'), response, options);
    }
    mockPUT(matcher, response, options) {
        return this.mock(utils_1.createRequestMatcher(matcher, 'PUT'), response, options);
    }
    mockDELETE(matcher, response, options) {
        return this.mock(utils_1.createRequestMatcher(matcher, 'DELETE'), response, options);
    }
    mockPATCH(matcher, response, options) {
        return this.mock(utils_1.createRequestMatcher(matcher, 'PATCH'), response, options);
    }
    removeMock(mock) {
        const index = this.mocks.indexOf(mock);
        if (index > -1) {
            return this.mocks.splice(index, 1)[0];
        }
    }
}
exports.Mockiavelli = Mockiavelli;
//# sourceMappingURL=mockiavelli.js.map